package mapx

import (
	"github.com/hsiafan/glow/v2/gox/option"
)

// FromKeys create map from keys
func FromKeys[K comparable, V any](keys []K, makeValue func(key K) V) map[K]V {
	var m = map[K]V{}
	for _, k := range keys {
		m[k] = makeValue(k)
	}
	return m
}

// FromValues create map from values
func FromValues[K comparable, V any](values []V, extractKey func(v V) K) map[K]V {
	var m = map[K]V{}
	for _, v := range values {
		m[extractKey(v)] = v
	}
	return m
}

// Copy shallow copy a map
func Copy[K comparable, V any](m map[K]V) map[K]V {
	if m == nil {
		return nil
	}
	var nm = map[K]V{}
	for k, v := range m {
		nm[k] = v
	}
	return nm
}

// ContainsKey return whether map contains the key
func ContainsKey[K comparable, V any](m map[K]V, k K) bool {
	_, ok := m[k]
	return ok
}

// ContainsAnyKey return whether map contains any one of the keys
func ContainsAnyKey[K comparable, V any](m map[K]V, keys ...K) bool {
	for _, k := range keys {
		if _, ok := m[k]; ok {
			return ok
		}
	}
	return false
}

// Get return map value of specified key, return default value if key not exists
func Get[K comparable, V any](m map[K]V, k K) option.Option[V] {
	if oldV, ok := m[k]; ok {
		return option.OfValue(oldV)
	}
	return option.Empty[V]()
}

// PutAll add all key-values entries into the map.
// If key already exists, will be set to new value.
func PutAll[K comparable, V any](m map[K]V, entries map[K]V) {
	for k, v := range entries {
		m[k] = v
	}
}

// PutIfAbsent add key-value to map, if key not exists; and return the current value.
func PutIfAbsent[K comparable, V any](m map[K]V, k K, v V) V {
	oldV, ok := m[k]
	if ok {
		return oldV
	}
	m[k] = v
	return v
}

// ComputeIfAbsent compute the value, add key-value to map, if key not exists; then return the current value.
func ComputeIfAbsent[K comparable, V any](m map[K]V, k K, makeValue func(key K) V) V {
	oldV, ok := m[k]
	if ok {
		return oldV
	}
	v := makeValue(k)
	m[k] = v
	return v
}

// ConvertValues create new map from original map, with values generated by original values and convert function.
func ConvertValues[K comparable, V any, U any](m map[K]V, convert func(key V) U) map[K]U {
	if m == nil {
		return nil
	}
	var newMap = map[K]U{}
	for k, v := range m {
		newMap[k] = convert(v)
	}
	return newMap
}

// Keys return map keys in a new Slice
func Keys[K comparable, V any](m map[K]V) []K {
	var s []K
	for v := range m {
		s = append(s, v)
	}
	return s
}

// Values return map value in a new slice
func Values[K comparable, V any](m map[K]V) []V {
	var values []V
	for _, v := range m {
		values = append(values, v)
	}
	return values
}

// Equals compare two maps, return true if hase same keys and values
func Equals[K comparable, V comparable](m1 map[K]V, m2 map[K]V) bool {
	if len(m1) != len(m2) {
		return false
	}
	for k, v := range m1 {
		v2, ok := m2[k]
		if !ok || v != v2 {
			return false
		}
	}
	return true
}
