package slicex

import (
	"errors"
	"sort"

	"github.com/hsiafan/glow/v2/gox/constraints"

	"github.com/hsiafan/glow/v2/gox/option"
)

// Copy copies a slice.
// copy can also trim slice underlying array to it's len.
func Copy[T any](s []T) []T {
	if s == nil {
		return nil
	}
	t := make([]T, len(s))
	copy(t, s)
	return t
}

// InvalidIndexErr slice index invalid
var InvalidIndexErr = errors.New("invalid index")

// Append add a new value to the end of slice.
func Append[T any](slice []T, v T) []T {
	return append(slice, v)
}

// Prepend add a new value to the beginning of slice.
func Prepend[T any](slice []T, v T) []T {
	var zero T
	slice = append(slice, zero)
	copy(slice[1:], slice)
	slice[0] = v
	return slice
}

// Concat concat multi slices into one slice
func Concat[T any](s1 []T, s2 []T, others ...[]T) []T {
	s := append(s1, s2...)
	for _, si := range others {
		s = append(s, si...)
	}
	return s
}

// Insert inserts new value at the position of slice.
// The index value must in range [0, len(slice)]
func Insert[T any](slice []T, index int, value T) []T {
	if index < 0 || index > len(slice) {
		panic(InvalidIndexErr)
	}
	if len(slice) == index {
		return append(slice, value)
	}
	slice = append(slice[:index+1], slice[index:]...)
	slice[index] = value
	return slice
}

// RemoveAt remove the value at the position of slice
// The index value must in range [0, len(slice)-1]
func RemoveAt[T any](slice []T, index int) []T {
	if index < 0 || index >= len(slice) {
		panic(InvalidIndexErr)
	}
	return append(slice[:index], slice[index+1:]...)
}

// RemoveBy remove the value match the given condition predicate.
func RemoveBy[T any](slice []T, predicate func(value T) bool) []T {
	var newIdx = 0
	for idx, v := range slice {
		if !predicate(v) {
			if idx != newIdx {
				slice[newIdx] = slice[idx]
			}
			newIdx++
		}
	}
	return slice[:newIdx]
}

// Remove removes the value match the given condition predicate.
func Remove[T comparable](slice []T, value T) []T {
	var newIdx = 0
	for idx, v := range slice {
		if v == value {
			if idx != newIdx {
				slice[newIdx] = slice[idx]
			}
			newIdx++
		}
	}
	return slice[:newIdx]
}

// Contains return whether slice contains the given value
func Contains[T comparable](s []T, value T) bool {
	for _, v := range s {
		if v == value {
			return true
		}
	}
	return false
}

// Find return the index of value; if not found, return -1
func Find[T comparable](s []T, value T) int {
	for idx, v := range s {
		if v == value {
			return idx
		}
	}
	return -1
}

// FindBy find the value match the predicate, return the index; if not found, return -1
func FindBy[T any](s []T, predicate func(v T) bool) int {
	for idx, v := range s {
		if predicate(v) {
			return idx
		}
	}
	return -1
}

// FindLast return the last index of lvalue; if not found, return -1
func FindLast[T comparable](s []T, value T) int {
	for idx := len(s) - 1; idx >= 0; idx-- {
		if s[idx] == value {
			return idx
		}
	}
	return -1
}

// First return the first element of slice in a gox.Option, return empty option if slice is empty.
func First[T comparable](s []T) option.Option[T] {
	if len(s) == 0 {
		return option.Empty[T]()
	}
	return option.OfValue(s[0])
}

// Last return the last element of slice in a gox.Option, return empty option if slice is empty.
func Last[T comparable](s []T) option.Option[T] {
	if len(s) == 0 {
		return option.Empty[T]()
	}
	return option.OfValue(s[len(s)-1])
}

// Select apply predicate on the origin slice, return a new slice contains the values match the predicate.
func Select[T any](s []T, predicate func(v T) bool) []T {
	var result []T
	for _, v := range s {
		if predicate(v) {
			result = append(result, v)
		}
	}
	return result
}

// Partition divide slice into two slice, one contains elements match the predicate, the other contains the mismatched.
func Partition[T any](s []T, predicate func(v T) bool) (matched []T, mismatched []T) {
	for _, v := range s {
		if predicate(v) {
			matched = append(matched, v)
		} else {
			mismatched = append(mismatched, v)
		}
	}
	return
}

// Convert apply convert func on the origin slice, return a new slice contains the generated value.
func Convert[T any, R any](s []T, convert func(v T) R) []R {
	result := make([]R, len(s))
	for idx, v := range s {
		result[idx] = convert(v)
	}
	return result
}

type orderedSlice[T constraints.Ordered] []T

func (o orderedSlice[T]) Len() int {
	return len(o)
}

func (o orderedSlice[T]) Less(i, j int) bool {
	return o[i] < o[j]
}

func (o orderedSlice[T]) Swap(i, j int) {
	o[i], o[j] = o[j], o[i]
}

// Sort sorts the slice in place
func Sort[T constraints.Ordered](s []T) {
	sort.Sort(orderedSlice[T](s))
}

// SortStable sorts the slice in place, with stable sort algorithm
func SortStable[T constraints.Ordered](s []T) {
	sort.Stable(orderedSlice[T](s))
}

// Sorted a new slice which contains elements from origin slice, and is sorted.
func Sorted[T constraints.Ordered](s []T) []T {
	slice := Copy(s)
	sort.Sort(orderedSlice[T](slice))
	return slice
}

type orderedSliceFunc[T any] struct {
	slice []T
	less  func(v1, v2 T) bool
}

func (o orderedSliceFunc[T]) Len() int {
	return len(o.slice)
}

func (o orderedSliceFunc[T]) Less(i, j int) bool {
	return o.less(o.slice[i], o.slice[j])
}

func (o orderedSliceFunc[T]) Swap(i, j int) {
	o.slice[i], o.slice[j] = o.slice[j], o.slice[i]
}

// SortBy sorts the slice in place, a less function is specified to compare slice values.
func SortBy[T any](s []T, less func(v1, v2 T) bool) {
	sort.Sort(orderedSliceFunc[T]{
		slice: s,
		less:  less,
	})
}

// SortStableBy sorts the slice in place, with stable sort algorithm. A less function is specified to compare slice values.
func SortStableBy[T any](s []T, less func(v1, v2 T) bool) {
	sort.Stable(orderedSliceFunc[T]{
		slice: s,
		less:  less,
	})
}

// Max returns an Option contains the max value in the slice. If slice is empty, return empty Option.
func Max[T constraints.Ordered](s []T) option.Option[T] {
	if len(s) == 0 {
		return option.Empty[T]()
	}
	max := s[0]
	for _, v := range s[1:] {
		if max < v {
			max = v
		}
	}
	return option.OfValue(max)
}

// MaxBy returns an Option contains the max value in the slice. If slice is empty, return empty Option.
// The values are compared by less func.
func MaxBy[T any](s []T, less func(v1, v2 T) bool) option.Option[T] {
	if len(s) == 0 {
		return option.Empty[T]()
	}
	max := s[0]
	for _, v := range s[1:] {
		if less(max, v) {
			max = v
		}
	}
	return option.OfValue(max)
}

// Min returns an Option contains the min value in the slice. If slice is empty, return empty Option.
func Min[T constraints.Ordered](s []T) option.Option[T] {
	if len(s) == 0 {
		return option.Empty[T]()
	}
	min := s[0]
	for _, v := range s[1:] {
		if min > v {
			min = v
		}
	}
	return option.OfValue(min)
}

// MinBy returns an Option contains the min value in the slice. If slice is empty, return empty Option.
// The values are compared by less func.
func MinBy[T any](s []T, less func(v1, v2 T) bool) option.Option[T] {
	if len(s) == 0 {
		return option.Empty[T]()
	}
	min := s[0]
	for _, v := range s[1:] {
		if less(v, min) {
			min = v
		}
	}
	return option.OfValue(min)
}

// Reduce returns max value in the slice. If slice is empty, return zero value.
func Reduce[T any, R any](s []T, initial R, reducer func(current R, value T) R) R {
	result := initial
	for _, v := range s {
		result = reducer(result, v)
	}
	return result
}
