package vector

import (
	"github.com/hsiafan/glow/v2/container/slicex"
	"github.com/hsiafan/glow/v2/gox/option"
)

// Vector is a slice but with mutable data-pointer/len/cap.
type Vector[T any] []T

// New create a new Vector. This func does defensive copy.
func New[T any](values ...T) Vector[T] {
	return slicex.Copy(values)
}

// NewWithCap create a new empty Vector with give capacity
func NewWithCap[T any](size int) Vector[T] {
	return make([]T, 0, size)
}

// Append add a new value to the end of Vector
func (v *Vector[T]) Append(value T) {
	s := []T(*v)
	*v = append(s, value)
}

// AppendAll add new values to the end of Vector
func (v *Vector[T]) AppendAll(value ...T) {
	s := []T(*v)
	*v = append(s, value...)
}

// Extend add new values in another Vector to the end of Vector
func (v *Vector[T]) Extend(vec Vector[T]) {
	s := []T(*v)
	*v = append(s, []T(vec)...)
}

// Insert inserts value at index
func (v *Vector[T]) Insert(index int, value T) {
	s := []T(*v)
	*v = slicex.Insert(s, index, value)
}

// RemoveAt removes value at index
func (v *Vector[T]) RemoveAt(index int) {
	s := []T(*v)
	*v = slicex.RemoveAt(s, index)
}

// RemoveBy removes all value match the given predicate
func (v *Vector[T]) RemoveBy(predicate func(value T) bool) {
	s := []T(*v)
	*v = slicex.RemoveBy(s, predicate)
}

// At returns element at index
func (v *Vector[T]) At(idx int) T {
	s := []T(*v)
	return s[idx]
}

// Set sets element at index
func (v *Vector[T]) Set(idx int, value T) {
	s := []T(*v)
	s[idx] = value
}

// SetSlice sets elements begin from index
func (v *Vector[T]) SetSlice(idx int, values []T) {
	s := []T(*v)
	copy(s[idx:], values)
}

// First returns the first element of Vector, in an Option.
// If Vector is empty, return empty Option
func (v *Vector[T]) First() option.Option[T] {
	s := []T(*v)
	if len(s) == 0 {
		return option.Empty[T]()
	}
	return option.OfValue(s[0])
}

// Last returns the last element of Vector, in an Option.
// If Vector is empty, return empty Option
func (v *Vector[T]) Last() option.Option[T] {
	s := []T(*v)
	if len(s) == 0 {
		return option.Empty[T]()
	}
	return option.OfValue(s[len(s)-1])
}

// RemoveLast remove and returns the last element of Vector, in an Option.
// If Vector is empty, return empty Option
func (v *Vector[T]) RemoveLast() option.Option[T] {
	s := []T(*v)
	if len(s) == 0 {
		return option.Empty[T]()
	}
	o := option.OfValue(s[len(s)-1])
	*v = s[:len(s)-1]
	return o
}

// Clear vector content, but do not reset capacity
func (v *Vector[T]) Clear() {
	s := []T(*v)
	*v = s[:0]
}

// Size returns elements count in Vector
func (v *Vector[T]) Size() int {
	s := []T(*v)
	return len(s)
}

// Empty returns if Vector has no element
func (v *Vector[T]) Empty() bool {
	s := []T(*v)
	return len(s) == 0
}

// Reserve requests that the vector capacity be at least enough to contain n elements
func (v *Vector[T]) Reserve(capacity int) {
	s := []T(*v)
	if cap(s) >= capacity {
		return
	}
	ns := make([]T, len(s), capacity)
	copy(ns, s)
	*v = ns
}

// ShrinkToFit requests the Vector to reduce its capacity to fit its size.
func (v *Vector[T]) ShrinkToFit() {
	s := []T(*v)
	if cap(s) > len(s) {
		ns := slicex.Copy(s)
		*v = ns
	}
}

// Capacity returns capacity of Vector
func (v *Vector[T]) Capacity() int {
	s := []T(*v)
	return cap(s)
}

// FindBy find first value match the given predicate, return the index.
// If not found, return -1
func (v *Vector[T]) FindBy(predicate func(v T) bool) int {
	s := []T(*v)
	return slicex.FindBy(s, predicate)
}

// Select return a new Vector contains the values match given predicate
func (v *Vector[T]) Select(predicate func(v T) bool) Vector[T] {
	s := []T(*v)
	return Vector[T](slicex.Select(s, predicate))
}

// SortBy sorts the vector in place, a less function is specified to compare slice values.
func (v *Vector[T]) SortBy(less func(v1, v2 T) bool) {
	s := []T(*v)
	slicex.SortBy(s, less)
}

// ForEach travers all elements in Vector
func (v *Vector[T]) ForEach(f func(v T)) {
	s := []T(*v)
	for _, e := range s {
		f(e)
	}
}

// SortStableBy sorts the vector in place, with stable sort algorithm. A less function is specified to compare slice values.
func (v *Vector[T]) SortStableBy(less func(v1, v2 T) bool) {
	s := []T(*v)
	slicex.SortStableBy(s, less)
}

// Copy return a new Vector contains same values as the origin Vector
func (v *Vector[T]) Copy() Vector[T] {
	s := []T(*v)
	t := make([]T, len(s))
	copy(t, s)
	return t
}

// AsSlice return the underlying slice, the slice share same data with Vector.
func (v *Vector[T]) AsSlice() []T {
	return *v
}

// ToSlice return a new slice, the slice contains a copy of data in Vector.
func (v *Vector[T]) ToSlice() []T {
	return slicex.Copy(*v)
}
