package vector

import (
	"testing"

	"github.com/stretchr/testify/assert"
)

func TestVector(t *testing.T) {
	var v Vector[int]
	v.Append(1)
	assert.Equal(t, []int{1}, v.AsSlice())

	v.Insert(0, 1)
	assert.Equal(t, []int{1, 1}, v.AsSlice())
	v.Insert(1, 2)
	assert.Equal(t, []int{1, 2, 1}, v.AsSlice())

	v.AppendAll(3, 4, 5)
	assert.Equal(t, []int{1, 2, 1, 3, 4, 5}, v.AsSlice())

	v.RemoveAt(2)
	assert.Equal(t, []int{1, 2, 3, 4, 5}, v.AsSlice())

	v.RemoveBy(func(value int) bool {
		return value < 2
	})
	assert.Equal(t, []int{2, 3, 4, 5}, v.AsSlice())

	v.Set(0, 3)
	assert.Equal(t, 3, v.At(0))
}

func TestVector_SortBy(t *testing.T) {
	v := New(1, 4, 3, 2)
	v.SortBy(func(v1, v2 int) bool {
		return v1 < v2
	})
	assert.Equal(t, []int{1, 2, 3, 4}, v.AsSlice())
}

func TestSelect(t *testing.T) {
	v := New(1, 2, 3)
	nv := v.Select(func(v int) bool {
		return v > 1
	})
	assert.Equal(t, []int{2, 3}, nv.AsSlice())
}

func TestVector_FindBy(t *testing.T) {
	v := New(1, 2, 3)
	idx := v.FindBy(func(v int) bool {
		return v > 1
	})
	assert.Equal(t, 1, idx)
	idx = v.FindBy(func(v int) bool {
		return v > 3
	})
	assert.Equal(t, -1, idx)
}

func TestVector_Clear(t *testing.T) {
	v := New(1, 2, 3)
	v.Clear()
	assert.Equal(t, 0, v.Size())
	assert.Equal(t, 3, v.Capacity())
	assert.Panics(t, func() {
		v.At(1)
	})
}

func TestVector_SetSlice(t *testing.T) {
	v := New(1, 2, 3)
	v.SetSlice(0, []int{3, 3})
	assert.Equal(t, []int{3, 3, 3}, v.AsSlice())
}

func TestVector_First(t *testing.T) {
	v := New(1, 2, 3)
	assert.Equal(t, 1, v.First().Must())
	v = New[int]()
	assert.False(t, v.First().Present())
}

func TestVector_Last(t *testing.T) {
	v := New(1, 2, 3)
	assert.Equal(t, 3, v.Last().Must())
	v = New[int]()
	assert.False(t, v.Last().Present())
}

func TestVector_Reserve(t *testing.T) {
	v := New(1, 2, 3)
	v.Reserve(16)
	v.Append(4)
	assert.Equal(t, []int{1, 2, 3, 4}, v.AsSlice())
}

func TestNewWithCap(t *testing.T) {
	v := NewWithCap[string](10)
	assert.Equal(t, 10, v.Capacity())
	assert.Equal(t, 0, v.Size())
}
