package flagx

import (
	"errors"
	"flag"
	"fmt"
	"os"
	"reflect"
	"strconv"
	"time"

	"github.com/hsiafan/glow/v2/mathx/floatx"
	"github.com/hsiafan/glow/v2/mathx/intx"
	"github.com/hsiafan/glow/v2/stringx"
)

const (
	flagTag        = "flag"        // mark this struct field as flag. The tag value is the flag name. If is empty, use converted(camel to low case, hyphen connected) struct filed name as flag name.
	defaultTag     = "default"     // set the default value for the flag. When defaultTag is specified, the arg is always treated as not required.
	descriptionTag = "description" // the usage message.
	requiredTag    = "required"    // If this args is required(true|false). Default is false. When default tag is set, the arg is always treated as not required.

	argumentTag = "argument" // mark this struct field to receive additional args.The value is the index of non-flag args, start from 0.If value is empty and this field is a slice, will receive all remained additional args.
)

// Command is a command line
type Command struct {
	Name                   string                 // the name of this command
	Description            string                 // usage message
	parentCmd              string                 // the composite command name, if this is a sub command
	flagSet                *flag.FlagSet          // for internal process
	flagFields             []*flagField           // flag field
	indexedArgumentFields  map[int]*argumentField // for storing indexed non-flag args field
	remainedArgumentsField *remainedArgsField     // for storing remained non-flag args
	handle                 func() error
}

var notStructErr = errors.New("not struct value")

// NewCommand create new command
func NewCommand[T any](name string, description string, handle func(option *T) error) *Command {
	flagSet := &flag.FlagSet{}
	var option = new(T)
	v := reflect.ValueOf(option)
	v = v.Elem()

	if v.Kind() != reflect.Struct {
		panic(notStructErr)
	}

	pr := &parseResult{
		argumentFields: map[int]*argumentField{},
	}

	for i := 0; i < v.NumField(); i++ {
		addOneField(v.Type().Field(i), v.Field(i), flagSet, pr)
	}

	// check if all non-flag args is consumed
	for idx := 0; idx < len(pr.argumentFields); idx++ {
		if _, ok := pr.argumentFields[idx]; !ok {
			panic(fmt.Errorf("no field receive %v-th non-flag arg", idx))
		}
	}

	cmd := &Command{
		flagSet:                flagSet,
		flagFields:             pr.fieldFlagValues,
		indexedArgumentFields:  pr.argumentFields,
		remainedArgumentsField: pr.remainArgumentField,
		Name:                   name,
		Description:            description,
		handle: func() error {
			return handle(option)
		},
	}

	flagSet.Usage = func() {
		output := flagSet.Output()
		if cmd.Description != "" {
			_, _ = fmt.Fprintln(output, cmd.Description+"\n")
		}

		argDes := argsDesc(pr.remainArgumentField, pr.argumentFields)
		if cmd.parentCmd != "" {
			_, _ = fmt.Fprintf(output, "Usage: %s %s %s\n", cmd.parentCmd, cmd.Name, argDes)
		} else {
			_, _ = fmt.Fprintf(output, "Usage: %s %s\n", cmd.Name, argDes)
		}

		flagSet.PrintDefaults()
	}
	return cmd

}

type parseResult struct {
	fieldFlagValues     []*flagField
	argumentFields      map[int]*argumentField
	remainArgumentField *remainedArgsField
}

func addOneField(sf reflect.StructField, fv reflect.Value, flagSet *flag.FlagSet, pr *parseResult) {
	if fv.Kind() == reflect.Struct && sf.Anonymous { //embed struct
		for i := 0; i < fv.NumField(); i++ {
			addOneField(fv.Type().Field(i), fv.Field(i), flagSet, pr)
		}
	}

	if !fv.IsValid() || !fv.CanSet() { // must be export fields
		return
	}

	if flagName, ok := sf.Tag.Lookup(flagTag); ok { // flag field
		if flagName == "" {
			flagName = toFlagName(sf.Name)
		}
		required := mustGetBoolTagValue(sf, requiredTag, false)
		defaultValue, hasDefault := sf.Tag.Lookup(defaultTag)
		var ffv = &flagField{
			name:         flagName,
			value:        fv,
			defaultValue: defaultValue,
			_type:        sf.Type,
			required:     required,
		}
		pr.fieldFlagValues = append(pr.fieldFlagValues, ffv)
		description := sf.Tag.Get(descriptionTag)
		if hasDefault {
			err := setValue(defaultValue, fv)
			if err != nil {
				panic(fmt.Errorf("invalid default value for field %v, error: %w", sf.Name, err))
			}
		}
		flagSet.Var(ffv, flagName, description)
	} else if indexStr, ok := sf.Tag.Lookup(argumentTag); ok {
		if indexStr != "" {
			index, err := strconv.Atoi(indexStr)
			if err != nil || index < 0 {
				panic(fmt.Errorf("invalid index value %v for field %v", indexStr, sf.Name))
			}

			if _, ok := pr.argumentFields[index]; ok {
				panic(fmt.Errorf("field %v and %v have same index %v", pr.argumentFields[index].fieldName, sf.Name, index))
			}

			pr.argumentFields[index] = &argumentField{
				value:     fv,
				_type:     sf.Type,
				fieldName: sf.Name,
				name:      toFlagName(sf.Name),
				index:     index,
			}
		} else {
			if sf.Type.Kind() != reflect.Slice {
				panic(fmt.Errorf("field %v want to receive all remains args, but is not slice type", sf.Name))
			}
			if pr.remainArgumentField != nil {
				panic(fmt.Errorf("both fields %v and %v want to receive remained non-flag args", pr.remainArgumentField.fieldName, sf.Name))
			} else {
				pr.remainArgumentField = &remainedArgsField{
					value:     fv,
					_type:     sf.Type,
					fieldName: sf.Name,
					name:      toFlagName(sf.Name),
				}
			}
		}
	}
}

// ParseOsArgsAndExecute parse commandline passed arguments, and run handlers.If error occurred, will exit with non-zero code.
func (c *Command) ParseOsArgsAndExecute() {
	c.ParseAndExecute(os.Args[1:])
}

// ParseAndExecute parse arguments, and run handlers. If error occurred, will exit with non-zero code.
func (c *Command) ParseAndExecute(arguments []string) {
	if err := c.flagSet.Parse(arguments); err != nil {
		if err == flag.ErrHelp {
			// already show usage
			return
		}
		c.exitOnError(fmt.Errorf("parse flag error: %w", err))
		return
	}

	for _, field := range c.flagFields {
		if field.required && !field.set {
			c.exitOnError(fmt.Errorf("flag %v required but not set", field.name))
		}
	}

	// deal with index arguments
	args := c.flagSet.Args()
	for i, arg := range args {
		if i >= len(c.indexedArgumentFields) {
			break
		}
		af := c.indexedArgumentFields[i]
		if err := af.Set(arg); err != nil {
			c.exitOnError(fmt.Errorf("set argument error: %w", err))
			return
		}
	}
	for _, field := range c.indexedArgumentFields {
		if !field.set {
			c.exitOnError(fmt.Errorf("arg %v[%v] required but not set", field.name, field.index))
		}
	}

	// deal with remained arguments
	if len(args) > len(c.indexedArgumentFields) {
		remainArgs := args[len(c.indexedArgumentFields):]
		if c.remainedArgumentsField == nil {
			c.exitOnError(fmt.Errorf("unrecognized args: %v", remainArgs))
		}
		if err := c.remainedArgumentsField.Set(remainArgs); err != nil {
			c.exitOnError(err)
			return
		}
	}

	if err := c.handle(); err != nil {
		c.exitOnError(err)
	}
}

// ShowUsage print formatted usage message
func (c *Command) ShowUsage() {
	c.flagSet.Usage()
}

func (c *Command) exitOnError(err error) {
	fmt.Println(err)
	c.ShowUsage()
	os.Exit(-1)
}

// make args description output for usage message
func argsDesc(raf *remainedArgsField, argFields map[int]*argumentField) string {
	var joiner = stringx.Joiner{
		Separator: " ",
	}
	var buffer = joiner.NewBuffer()

	for idx := 0; idx < len(argFields); idx++ {
		f := argFields[idx]
		buffer.Add(f.name)
	}

	if raf != nil {
		buffer.Add("[" + raf.name + "...]")
	}
	return buffer.String()
}

var _ flag.Value = (*flagField)(nil)

// flag.Value implementation that store value in a struct field
type flagField struct {
	name         string
	defaultValue string
	_type        reflect.Type
	value        reflect.Value
	required     bool
	set          bool
}

func (f *flagField) String() string {
	return f.defaultValue
}

func (f *flagField) Set(s string) error {
	if f.set && f._type.Kind() != reflect.Slice {
		return errors.New("flag arg " + f.name + " already set")
	}
	f.set = true
	if f._type.Kind() != reflect.Slice {
		return setValue(s, f.value)
	}
	newValue := reflect.New(f._type.Elem())
	if err := setValue(s, newValue.Elem()); err != nil {
		return fmt.Errorf("set %v argument error: %w", f.name, err)
	}
	f.value.Set(reflect.Append(f.value, newValue.Elem()))
	return nil
}

func (f *flagField) IsBoolFlag() bool {
	return f._type.Kind() == reflect.Bool
}

// for remember non-flag positional args.
type argumentField struct {
	value     reflect.Value
	_type     reflect.Type
	fieldName string // struct field name
	name      string // flag name
	index     int    // non-flag arg index
	set       bool
}

func (f *argumentField) Set(s string) error {
	f.set = true
	return setValue(s, f.value)
}

// for remember non-flag remained args.
type remainedArgsField struct {
	value     reflect.Value
	_type     reflect.Type
	fieldName string // struct field name
	name      string // flag name
}

func (f *remainedArgsField) Set(args []string) error {
	slice := reflect.MakeSlice(f._type, len(args), len(args))
	for idx, arg := range args {
		if err := setValue(arg, slice.Index(idx)); err != nil {
			return fmt.Errorf("set remained arguments error: %w", err)
		}
	}
	f.value.Set(slice)
	return nil
}

// set one arg field, the value converted from str input
func setValue(str string, value reflect.Value) error {
	switch value.Kind() {
	case reflect.String:
		value.SetString(str)
	case reflect.Int:
		v, err := intx.Parse[int](str)
		if err != nil {
			return err
		}
		value.SetInt(int64(v))
	case reflect.Int8:
		v, err := intx.Parse[int8](str)
		if err != nil {
			return err
		}
		value.SetInt(int64(v))
	case reflect.Int16:
		v, err := intx.Parse[int16](str)
		if err != nil {
			return err
		}
		value.SetInt(int64(v))
	case reflect.Int32:
		v, err := intx.Parse[int32](str)
		if err != nil {
			return err
		}
		value.SetInt(int64(v))
	case reflect.Int64:
		switch value.Interface().(type) {
		case time.Duration:
			v, err := time.ParseDuration(str)
			if err != nil {
				return err
			}
			value.Set(reflect.ValueOf(v))
			return nil
		}
		v, err := intx.Parse[int64](str)
		if err != nil {
			return err
		}
		value.SetInt(v)
	case reflect.Uint:
		v, err := intx.ParseUnsigned[uint](str)
		if err != nil {
			return err
		}
		value.SetUint(uint64(v))
	case reflect.Uint8:
		v, err := intx.ParseUnsigned[uint8](str)
		if err != nil {
			return err
		}
		value.SetUint(uint64(v))
	case reflect.Uint16:
		v, err := intx.ParseUnsigned[uint16](str)
		if err != nil {
			return err
		}
		value.SetUint(uint64(v))
	case reflect.Uint32:
		v, err := intx.ParseUnsigned[uint32](str)
		if err != nil {
			return err
		}
		value.SetUint(uint64(v))
	case reflect.Uint64:
		v, err := intx.ParseUnsigned[uint64](str)
		if err != nil {
			return err
		}
		value.SetUint(v)
	case reflect.Float32:
		v, err := floatx.Parse[float32](str)
		if err != nil {
			return err
		}
		value.SetFloat(float64(v))
	case reflect.Float64:
		v, err := floatx.Parse[float64](str)
		if err != nil {
			return err
		}
		value.SetFloat(v)
	case reflect.Bool:
		v, err := strconv.ParseBool(str)
		if err != nil {
			return err
		}
		value.SetBool(v)
	default:
		return fmt.Errorf("unsupported field type: %v", value.Kind())
	}
	return nil
}
