package lazy

import (
	"sync"
	"sync/atomic"
)

type noCopy struct{}

func (*noCopy) Lock()   {}
func (*noCopy) Unlock() {}

// Lazy for lazy init a value.
// The value is computed only after Get method is called, and can only compute successfully once.
// If compute value panics, value will be computed again next time.
type Lazy[T any] struct {
	noCopy  noCopy
	done    int32
	value   T
	compute func() T
	lock    sync.Mutex
}

// Of create a new Lazy instance
func Of[T any](compute func() T) *Lazy[T] {
	return &Lazy[T]{
		compute: compute,
	}
}

// Get return the lazy value. If value is not present yet, compute it.
// If value is already computing by another goroutine, wait for computing to finished.
// If panics while computing the value, this method panics
func (l *Lazy[T]) Get() T {
	if atomic.LoadInt32(&l.done) == 1 {
		return l.value
	}
	l.lock.Lock()
	defer l.lock.Unlock()
	if l.done == 0 {
		l.value = l.compute()
		atomic.StoreInt32(&l.done, 1)
	}

	return l.value
}

// TryGet return the lazy value. If value is not present yet, return Option.
// This method will not compute the value when the value not exits.
func (l *Lazy[T]) TryGet() (value T, exits bool) {
	if atomic.LoadInt32(&l.done) == 1 {
		return l.value, true
	}
	return
}
