package option

import "errors"

// Option manages an option contained value, i.e. a value that may or may not be present.
type Option[T any] struct {
	value   T
	present bool
}

// Of create an Option
func Of[T any](value T, present bool) Option[T] {
	return Option[T]{
		value:   value,
		present: present,
	}
}

// OfValue create a non-empty Option, with a value
func OfValue[T any](value T) Option[T] {
	return Option[T]{
		value:   value,
		present: true,
	}
}

// Empty returns an empty Option
func Empty[T any]() Option[T] {
	return Option[T]{}
}

// Unpack returns the state of option
func (o Option[T]) Unpack() (value T, present bool) {
	return o.value, o.present
}

// Present returns if option value is present
func (o Option[T]) Present() bool {
	return o.present
}

var valueNotPresentErr = errors.New("option value not present")

// Must return the value Option contains. If value is not present, panics
func (o Option[T]) Must() T {
	if !o.present {
		panic(valueNotPresentErr)
	}
	return o.value
}

// OrElse return the value Option contains. If value is not present, return a default value.
func (o Option[T]) OrElse(defaultValue T) T {
	if !o.present {
		return defaultValue
	}
	return o.value
}

// OrCompute return the value Option contains. If value is not present, return a value computed by the func.
func (o Option[T]) OrCompute(f func() T) T {
	if !o.present {
		return f()
	}
	return o.value
}

// OrNext return the Option if it is present. If value is not present, return an Option computed by the func.
func (o Option[T]) OrNext(next func() Option[T]) Option[T] {
	if !o.present {
		return next()
	}
	return o
}
