package filex

import (
	"os"

	"github.com/hsiafan/glow/v2/iox"
	"github.com/hsiafan/glow/v2/stringx"
	"golang.org/x/text/encoding"
)

// Create creates a file for writing, truncate the file if file already exists.
func Create(filename string) (*os.File, error) {
	return os.Create(filename)
}

// CreateOrAppend create a file for writing, append to file if file already exists.
func CreateOrAppend(filename string) (*os.File, error) {
	return os.OpenFile(filename, os.O_APPEND|os.O_WRONLY|os.O_CREATE, 0600)
}

// ReadAllToStringWithEncoding read and return all data as string in file
func ReadAllToStringWithEncoding(path string, enc encoding.Encoding) (string, error) {
	data, err := os.ReadFile(path)
	if err != nil {
		return "", err
	}
	decoded, err := enc.NewDecoder().Bytes(data)
	if err != nil {
		return "", err
	}
	return string(decoded), err
}

// ReadAllToString read and return all data as string in file
func ReadAllToString(path string) (string, error) {
	data, err := os.ReadFile(path)
	if err != nil {
		return "", err
	}
	return stringx.WrapBytes(data), err
}

// ReadAllBytes read and return all data in file
func ReadAllBytes(path string) ([]byte, error) {
	return os.ReadFile(path)
}

// ForEachLineWithEncoding read line by line from a file with specific, and return a error if read failed.
func ForEachLineWithEncoding(path string, enc encoding.Encoding, consume func(line string)) error {
	reader, err := os.Open(path)
	if err != nil {
		return err
	}
	defer iox.Close(reader)
	return iox.ForEachLineWithEncoding(reader, enc, consume)
}

// ForEachLine read line by line from a file, and return a error if read failed.
func ForEachLine(path string, consume func(line string)) error {
	reader, err := os.Open(path)
	if err != nil {
		return err
	}
	defer iox.Close(reader)
	return iox.ForEachLine(reader, consume)
}

// ReadAllLinesWithEncoding read all data from a file till EOF, return a lines slice.
func ReadAllLinesWithEncoding(path string, enc encoding.Encoding) ([]string, error) {
	reader, err := os.Open(path)
	if err != nil {
		return nil, err
	}
	defer iox.Close(reader)
	return iox.ReadAllLinesWithEncoding(reader, enc)
}

// ReadAllLines read all data from a file till EOF, return a lines slice.
func ReadAllLines(path string) ([]string, error) {
	reader, err := os.Open(path)
	if err != nil {
		return nil, err
	}
	defer iox.Close(reader)
	return iox.ReadAllLines(reader)
}

// Exists check if file exists
func Exists(path string) (bool, error) {
	_, err := os.Stat(path)
	if err == nil {
		return true, nil
	}
	if os.IsNotExist(err) {
		return false, nil
	}
	return false, err
}

// IsFile return true if path is exists and is regular file
func IsFile(path string) (bool, error) {
	fi, err := os.Stat(path)
	if err != nil {
		if os.IsNotExist(err) {
			return false, nil
		}
		return false, err
	}
	switch mode := fi.Mode(); {
	case mode.IsRegular():
		return true, nil
	default:
		return false, nil
	}
}

// IsDirectory return true if path is exists and is directory
func IsDirectory(path string) (bool, error) {
	fi, err := os.Stat(path)
	if err != nil {
		if os.IsNotExist(err) {
			return false, nil
		}
		return false, err
	}
	switch mode := fi.Mode(); {
	case mode.IsDir():
		return true, nil
	default:
		return false, nil
	}
}

// WriteBytes write all data to file, and then close. If file already exists, will be overridden
func WriteBytes(path string, data []byte) error {
	fi, err := os.Create(path)
	if err != nil {
		return err
	}
	defer iox.Close(fi)
	_, err = fi.Write(data)
	return err
}

// WriteString write all string content to file, and then close. If file already exists, will be overridden
func WriteString(path string, str string) error {
	return WriteBytes(path, stringx.AsBytes(str))
}

// WriteStringWithEncoding write all string content to file using specific encoding, and then close.
// If file already exists, will be overridden
func WriteStringWithEncoding(path string, str string, enc encoding.Encoding) error {
	data, err := enc.NewEncoder().Bytes(stringx.AsBytes(str))
	if err != nil {
		return err
	}
	return WriteBytes(path, data)
}

// MkDirs creates a directory named path, along with any necessary parents. The umask of dirs created is set to 0o777.
// Return err is any err occurred.If path is already a directory, MkdirAll does nothing and returns nil.
func MkDirs(path string) error {
	return os.MkdirAll(path, os.ModePerm)
}
