package reflectx

import (
	"errors"
	"reflect"
	"unsafe"

	"github.com/hsiafan/glow/v2/stringx"
)

var notStructErr = errors.New("not struct value")

// valueStruct has same layout as reflect.Value struct
type valueStruct struct {
	typ  unsafe.Pointer
	ptr  unsafe.Pointer
	flag uintptr
}

const (
	flagStickyRO    = 1 << 5 // obtained via unexported not embedded field, so read-only
	flagEmbedRO     = 1 << 6 // obtained via unexported embedded field, so read-only
	flagIndir       = 1 << 7 // val holds a pointer to the data
	flagAddr        = 1 << 8 // v.CanAddr is true (implies flagIndir)
	flagMethod      = 1 << 9 // a method value
	flagMethodShift = 10
	flagRO          = flagStickyRO | flagEmbedRO
)

// TypeOf return reflect.Type for T
func TypeOf[T any]() reflect.Type {
	return reflect.TypeOf((*T)(nil)).Elem()
}

// IsBytesType return if t is type for []byte
func IsBytesType(t reflect.Type) bool {
	return t.Kind() == reflect.Slice && t.Elem().Kind() == reflect.Uint8
}

// ForceGetValue force get the value hold by reflect.Value.
func ForceGetValue[T any](value reflect.Value) T {
	var zero T
	if unsafe.Sizeof(zero) != value.Type().Size() {
		panic("types have different size")
	}
	if value.Type().Kind() == reflect.Func {
		panic("func value not supported")
	}
	vsp := (*valueStruct)(unsafe.Pointer(&value))

	if vsp.flag&flagIndir == 0 {
		return *(*T)(unsafe.Pointer(&vsp.ptr))
	} else {
		return *(*T)(vsp.ptr)
	}
}

// ForceSetValue force set reflect.Value, no matter where value if readonly, not addressable, or has different types(but has same size).
func ForceSetValue[T any](value *reflect.Value, v T) {
	if unsafe.Sizeof(v) != value.Type().Size() {
		panic("force store type has different size")
	}
	if value.Type().Kind() == reflect.Func {
		panic("func value not supported")
	}

	vsp := (*valueStruct)(unsafe.Pointer(value))

	if vsp.flag&flagIndir == 0 {
		*(*T)(unsafe.Pointer(&vsp.ptr)) = v
	} else {
		if vsp.ptr == nil || vsp.flag&flagAddr == 0 {
			vsp.ptr = unsafe.Pointer(&v)
		} else {
			*(*T)(vsp.ptr) = v
		}
	}
}

// StructToMap shallow convert struct to map, using field name as key, field value as map value.
// The Field name can be overridden using 'name' struct tag.
//
// Note: only exported fields are set into map.
func StructToMap(value any) map[string]any {
	v := reflect.ValueOf(value)
	switch v.Kind() {
	case reflect.Ptr:
		if v.IsNil() {
			return nil
		}
		return toMap(v.Elem())
	case reflect.Struct:
		return toMap(v)
	default:
		panic(notStructErr)
	}
}

func toMap(v reflect.Value) map[string]any {
	if v.Kind() != reflect.Struct {
		panic(notStructErr)
	}
	m := make(map[string]interface{})
	for i := 0; i < v.NumField(); i++ {
		fv := v.Field(i)
		sf := v.Type().Field(i)

		if fv.Kind() == reflect.Ptr {
			fv = fv.Elem()
		}
		if fv.Kind() == reflect.Struct && sf.Anonymous { // embed struct
			for k, v := range toMap(fv) {
				if _, ok := m[k]; !ok {
					m[k] = v
				}
			}
		} else {
			if fv.CanInterface() { // true when is exported field
				name := stringx.FirstNonEmpty(sf.Tag.Get("name"), sf.Name)
				m[name] = fv.Interface()
			}
		}

	}
	return m
}
