package reflectx

import (
	"errors"
	"fmt"
	"reflect"
	"testing"
	"unsafe"

	"github.com/stretchr/testify/assert"
)

func TestToMap(t *testing.T) {
	var s = struct {
		Test  string
		Test2 int
		test3 int64
	}{
		Test:  "1",
		Test2: 2,
		test3: 10,
	}
	assert.Equal(t, map[string]any{"Test": "1", "Test2": 2}, StructToMap(s))
	assert.Equal(t, map[string]any{"Test": "1", "Test2": 2}, StructToMap(&s))
}

func TestToMapEmbedded(t *testing.T) {
	var s Test
	s.Value = 1
	s.Test2 = &Test2{
		Value2: "2",
	}
	assert.Equal(t, map[string]any{"Value": 1, "Value1": 0, "Value2": "2", "Value3": ""}, StructToMap(s))
	assert.Equal(t, map[string]any{"Value": 1, "Value1": 0, "Value2": "2", "Value3": ""}, StructToMap(&s))
}

type Test struct {
	Value int
	Test1
	test3
}

type Test1 struct {
	Value1 int
	*Test2
}

type Test2 struct {
	Value2 string
	Value  int
}

type test3 struct {
	Value3 string
}

func TestTypeOf(t *testing.T) {
	assert.Equal(t, reflect.TypeOf(1), TypeOf[int]())
	var i fmt.Stringer = nil
	assert.Equal(t, reflect.TypeOf(&i).Elem(), TypeOf[fmt.Stringer]())
	assert.Equal(t, reflect.TypeOf(&i), TypeOf[*fmt.Stringer]())
}

func TestIsBytes(t *testing.T) {
	var bs []byte
	assert.True(t, IsBytesType(reflect.TypeOf(bs)))
	assert.False(t, IsBytesType(reflect.TypeOf(1)))
}

func TestForceSetValue(t *testing.T) {
	// test not addressable value interleave. may have same pointer address for different values
	{
		var j = 1
		var ii = 1
		vv := reflect.ValueOf(j)
		ForceSetValue(&vv, 2)
		vv2 := reflect.ValueOf(ii)
		ForceSetValue(&vv2, 3)
	}

	// byte
	v := reflect.ValueOf(byte(1))
	ForceSetValue(&v, byte(2))
	assert.Equal(t, byte(2), byte(v.Uint()))

	// int/string maybe interned
	v = reflect.ValueOf(1)
	ForceSetValue(&v, 2)
	assert.Equal(t, 2, int(v.Int()))

	v = reflect.ValueOf("test")
	ForceSetValue(&v, "test2")
	assert.Equal(t, "test2", v.String())

	// struct fit into an uintptr, should use direct store
	var s = "1234"
	type holder struct {
		pointer unsafe.Pointer
	}
	h := holder{}
	hv := reflect.ValueOf(h)
	ForceSetValue(&hv, holder{pointer: unsafe.Pointer(&s)})
	assert.Equal(t, unsafe.Pointer(&s), hv.Interface().(holder).pointer)

	type small struct {
		b byte
	}
	ss := small{}
	ssv := reflect.ValueOf(ss)
	ForceSetValue(&ssv, small{b: 12})
	assert.Equal(t, byte(12), ssv.Interface().(small).b)

	// struct field value test
	type innerStruct struct {
		value int
		name  string
		holder
	}
	sp := &innerStruct{
		value: 1,
		name:  "test",
	}
	sv := reflect.ValueOf(sp)
	nameFV := sv.Elem().FieldByName("name")
	ForceSetValue(&nameFV, "test2")
	assert.Equal(t, "test2", nameFV.String())
	assert.Equal(t, "test2", sp.name)
	holderFV := sv.Elem().FieldByName("holder")
	ForceSetValue(&holderFV, holder{pointer: unsafe.Pointer(&s)})
	assert.Equal(t, unsafe.Pointer(&s), sp.holder.pointer)

	// map
	v = reflect.ValueOf(map[string]int{"1": 1})
	ForceSetValue(&v, map[string]int{"1": 2})
	assert.Equal(t, map[string]int{"1": 2}, v.Interface().(map[string]int))

	// pointer
	var i = 10
	var j = 11
	var ip = &i
	var jp = &j
	v = reflect.ValueOf(ip)
	ForceSetValue(&v, jp)
	assert.Equal(t, jp, v.Interface().(*int))
	ev := v.Elem()
	ForceSetValue(&ev, 12)
	assert.Equal(t, 12, *jp)
	assert.Equal(t, 10, *ip)

	// interface
	var it = errors.New("test")
	var it2 = errors.New("test2")
	v = reflect.ValueOf(&it).Elem()
	ForceSetValue(&v, it2)
	assert.Equal(t, it2, v.Interface().(error))
}

func TestForceGetValue(t *testing.T) {
	v := reflect.ValueOf(byte(1))
	assert.Equal(t, byte(1), ForceGetValue[byte](v))

	v = reflect.ValueOf(1)
	assert.Equal(t, 1, ForceGetValue[int](v))

	v = reflect.ValueOf("test")
	assert.Equal(t, "test", ForceGetValue[string](v))

	var s = "1234"
	type holder struct {
		pointer unsafe.Pointer
	}
	h := holder{pointer: unsafe.Pointer(&s)}
	v = reflect.ValueOf(h)
	assert.Equal(t, h, ForceGetValue[holder](v))

	type small struct {
		b byte
	}
	ss := small{b: 12}
	v = reflect.ValueOf(ss)
	assert.Equal(t, ss, ForceGetValue[small](v))

	v = reflect.ValueOf(map[string]int{"1": 1})
	assert.Equal(t, map[string]int{"1": 1}, ForceGetValue[map[string]int](v))

	var i = 10
	var ip = &i
	v = reflect.ValueOf(ip)
	assert.Equal(t, ip, ForceGetValue[*int](v))
	assert.Equal(t, 10, ForceGetValue[int](v.Elem()))

	var it = errors.New("test")
	v = reflect.ValueOf(&it).Elem()
	assert.Equal(t, it, ForceGetValue[error](v))
}
