package timex

import (
	"errors"
	"time"
)

type noCopy struct{}

func (*noCopy) Lock()   {}
func (*noCopy) Unlock() {}

// Stopwatch for measure elapsed time.
// The zero value is valid, as an un-started stopwatch.
type Stopwatch struct {
	noCopy      noCopy
	startedTime time.Time     // last start time
	elapsed     time.Duration // total elapsed time till last start
	running     bool          // if stopwatch is running
}

// NewStopwatch create new Stopwatch.
func NewStopwatch() *Stopwatch {
	return &Stopwatch{}
}

// NewStopwatchStarted create new Stopwatch, then start it.
func NewStopwatchStarted() *Stopwatch {
	return &Stopwatch{
		startedTime: time.Now(),
		running:     true,
	}
}

// Start the Stopwatch. If StopWatch is already started, panics with an error.
// A stopped Stopwatch can start again.
func (w *Stopwatch) Start() {
	if w.running {
		panic(errors.New("stopwatch already started"))
	}
	w.running = true
	w.startedTime = time.Now()
}

// Stop the Stopwatch. The elapsed time will not when stopwatch is stopped.
// If Stopwatch is not started, panics with an error.
func (w *Stopwatch) Stop() {
	if !w.running {
		panic(errors.New("stopwatch not started"))
	}
	w.running = false
	w.elapsed = w.elapsed + time.Now().Sub(w.startedTime)
}

// Reset resets the Stopwatch. The elapsed time will set to zero, the status will set to not running.
func (w *Stopwatch) Reset() {
	w.elapsed = 0
	w.running = false
}

// Elapsed return elapsed time.
func (w *Stopwatch) Elapsed() time.Duration {
	if !w.running {
		return w.elapsed
	}
	return w.elapsed + time.Now().Sub(w.startedTime)
}

// ElapsedMillis return elapsed time by milliseconds.
func (w *Stopwatch) ElapsedMillis() int64 {
	return int64(w.Elapsed() / time.Millisecond)
}

// ElapsedNanos return elapsed time by nanoseconds.
func (w *Stopwatch) ElapsedNanos() int64 {
	return int64(w.Elapsed() / time.Nanosecond)
}
